# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import dataclasses
import pathlib
import typing

from kitsunekko_tools.common import KitsuException
from kitsunekko_tools.config import KitsuConfig
from kitsunekko_tools.consts import IGNORE_FILENAME


@dataclasses.dataclass(frozen=True)
class IgnoreListException(KitsuException):
    what: str


class IgnoreList:
    """
    Holds a list of files that should not be downloaded even if they're not present in expected locations.
    """

    _config: KitsuConfig
    _ignore_filepath: pathlib.Path
    _patterns: dict[str, None]
    _dirty_level: int  # counts additions
    _autocommit_threshold: int

    def __init__(self, config: KitsuConfig, autocommit_threshold: int = 20):
        self._config = config
        self._ignore_filepath = pathlib.Path(self._config.destination) / IGNORE_FILENAME
        self._patterns = {}
        self._dirty_level = 0
        self._autocommit_threshold = autocommit_threshold
        self._config.raise_for_destination()
        try:
            with open(self._ignore_filepath, encoding="utf8") as f:
                self._patterns.update(dict.fromkeys(filter(bool, map(str.strip, f.read().splitlines()))))
        except FileNotFoundError:
            pass

    @property
    def ignore_filepath(self) -> pathlib.Path:
        return self._ignore_filepath

    def _pattern_from_path(self, file_path: pathlib.Path) -> str:
        return str(file_path.relative_to(self._config.destination))

    def is_matching(self, file_path: pathlib.Path) -> bool:
        return self._pattern_from_path(file_path) in self._patterns

    def patterns(self) -> typing.Iterable[str]:
        """
        Return all known ignore patterns.
        """
        return self._patterns.keys()

    def add(self, pattern: str) -> None:
        """
        Add a new ignore pattern to the list.
        """
        self._patterns[pattern] = None
        self._dirty_level += 1

    def add_file(self, file_path: pathlib.Path) -> None:
        """
        Add file to the list, as relative path.
        """
        return self.add(self._pattern_from_path(file_path))

    def path(self) -> pathlib.Path:
        """
        Return path to ignore file.
        """
        return self._ignore_filepath

    def commit(self) -> None:
        """
        Save ignore file to disk.
        """
        if self._dirty_level == 0:
            return
        data = "\n".join(self._patterns) + "\n"
        self._ignore_filepath.write_text(data, encoding="utf-8")  # newline at the end of file
        self._dirty_level = 0

    def maybe_commit_midway(self) -> None:
        """
        Save file to disk if its dirty level exceeds the set limit.
        """
        if self._dirty_level >= self._autocommit_threshold:
            self.commit()



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import abc
import re


class KitsuException(Exception, abc.ABC):
    @property
    @abc.abstractmethod
    def what(self) -> str:
        raise NotImplementedError()


RE_FILENAME_PROHIBITED = re.compile(r"[ _\\\n\t\r#{}<>^*/:\"`?'|]+", flags=re.MULTILINE | re.IGNORECASE)
RE_MULTI_SPACE = re.compile(r" {2,}", flags=re.MULTILINE | re.IGNORECASE)
WINDOWS_SUBSTITUTE_CHARS = {
    "??": "2",
    "||": "2",
    "'": "’",
    "\\": "⧵",
    "/": "∕",
    ":": ".",
    "?": "？",
    "|": "⏐",
}
assert all(k != v for k, v in WINDOWS_SUBSTITUTE_CHARS.items())


def fs_name_strip(name: str) -> str:
    for from_, to in WINDOWS_SUBSTITUTE_CHARS.items():
        name = name.replace(from_, to)
    name = re.sub(RE_FILENAME_PROHIBITED, " ", name)
    name = re.sub(RE_MULTI_SPACE, " ", name)
    return name.strip()



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
import datetime
import typing


class CompareError(ValueError):
    def __init__(self, obj1, obj2):
        super().__init__(f"Can't compare type {type(obj1).__name__} to type {type(obj2).__name__}")


class AnimeDir(typing.NamedTuple):
    url: str  # full URL to the directory with subtitle files
    show_name: str  # name of the anime
    mod_timestamp: datetime.datetime  # last modified

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AnimeDir):
            raise CompareError(self, other)
        return self.show_name == other.show_name

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.show_name)


class SubtitleFile(typing.NamedTuple):
    url: str  # full URL to the subtitle file
    show_name: str  # anime title
    file_name: str  # name of the subtitle file
    mod_timestamp: datetime.datetime  # last modified

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SubtitleFile):
            raise CompareError(self, other)
        return self.show_name == other.show_name and self.file_name == other.file_name

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(f"{self.show_name}/{self.file_name}")



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import datetime
import itertools
import pathlib
import re
import typing
import urllib.parse

from kitsunekko_tools.common import fs_name_strip
from kitsunekko_tools.consts import KITSUNEKKO_DOMAIN_URL
from kitsunekko_tools.scrapper.types import AnimeDir, SubtitleFile

MOD_TIMESTAMP_FORMAT = "%b %d %Y %I:%M:%S %p"  # timestamp format used on kitsunekko


def datetime_from_str(mod_timestamp: str) -> datetime.datetime:
    return datetime.datetime.strptime(mod_timestamp, MOD_TIMESTAMP_FORMAT)


RE_FLAGS = re.IGNORECASE

# <tr><td colspan="2"><a href="/dirlist.php?dir=subtitles%2Fjapanese%2FMushoku+Tensei%3A+Isekai+Ittara+Honki+Dasu%2F" class=""><strong>Mushoku Tensei: Isekai Ittara Honki Dasu</strong> </a></td> <td class="tdright" title="Dec 03 2023 05:13:17 AM" > 4&nbsp;months </td></tr>
# <tr><td colspan="2"><a href="/dirlist.php?dir=subtitles%2Fjapanese%2FSousou+no+Frieren%2F" class=""><strong>Sousou no Frieren</strong> </a></td> <td class="tdright" title="Apr 14 2024 06:23:19 PM" > 1&nbsp;week </td></tr>
RE_SUBTITLE_DIR = re.compile(
    r'<a href="/?(?P<abs_path>dirlist.php\?dir=[^"\']+)"[^<>]*>\s*<strong>\s*(?P<show_name>.+?)\s*</strong>\s*</a>.*<td class="tdright" title="(?P<mod_timestamp>[^<>"]+)"\s*>',
    flags=RE_FLAGS,
)

# <tr><td><a href="subtitles/japanese/Henjin no Salad Bowl/Henjin no Salad Bowl - 01 「麒麟がくる(異世界から)」 (TBS 1920x1080 x264 AAC).srt" class=""><strong>Henjin no Salad Bowl - 01 「麒麟がくる(異世界から)」 (TBS 1920x1080 x264 AAC).srt</strong> </a></td> <td class="tdleft"  title="29823"  > 29&nbsp;KB </td> <td class="tdright" title="Apr 05 2024 05:45:25 AM" > 3&nbsp;weeks </td></tr>
# <tr><td><a href="subtitles/japanese/Henjin no Salad Bowl/Henjin no Salad Bowl - 02 「ホームレス女騎士／はじめてのおしごと他」 (TBS 1920x1080 x264 AAC).srt" class=""><strong>Henjin no Salad Bowl - 02 「ホームレス女騎士／はじめてのおしごと他」 (TBS 1920x1080 x264 AAC).srt</strong> </a></td> <td class="tdleft"  title="31494"  > 31&nbsp;KB </td> <td class="tdright" title="Apr 12 2024 01:28:54 PM" > 2&nbsp;weeks </td></tr>
RE_SUBTITLE_FILE = re.compile(
    r'<a href="/?(?P<abs_path>subtitles/[^"\']+\.(?:zip|rar|7z|ass|srt|ssa))"[^<>]*>.*<td class="tdright" title="(?P<mod_timestamp>[^<>"]+)"\s*>',
    flags=RE_FLAGS,
)


def sanitize_name(title: str) -> str:
    return fs_name_strip(urllib.parse.unquote(title))


def find_all_subtitle_dirs(html_text: str) -> typing.Iterable[AnimeDir]:
    for match in re.finditer(RE_SUBTITLE_DIR, html_text):
        yield AnimeDir(
            url=f"{KITSUNEKKO_DOMAIN_URL}/{match.group('abs_path')}",
            show_name=sanitize_name(match.group("show_name")),
            # timestamp input looks like "Jul 15 2012 09:24:15 PM"
            mod_timestamp=datetime_from_str(match.group("mod_timestamp").strip()),
        )


def find_all_subtitle_files(html_text: str) -> typing.Iterable[SubtitleFile]:
    for match in re.finditer(RE_SUBTITLE_FILE, html_text):
        show_name, file_name = match.group("abs_path").split("/")[-2:]
        yield SubtitleFile(
            url=f"{KITSUNEKKO_DOMAIN_URL}/{urllib.parse.quote(match.group('abs_path'))}",
            show_name=sanitize_name(show_name),
            file_name=sanitize_name(file_name),
            # timestamp input looks like "Jul 15 2012 09:24:15 PM"
            mod_timestamp=datetime_from_str(match.group("mod_timestamp").strip()),
        )


def main():
    from pprint import pprint

    data_dir = pathlib.Path(__file__).parent.joinpath("../tests/data")

    html_text = data_dir.joinpath("main_dir_page.html").read_text()
    for anime_dir in itertools.islice(find_all_subtitle_dirs(html_text), 10):
        pprint(anime_dir, indent=2)

    html_text = data_dir.joinpath("subs_page.html").read_text()
    for sub_file in itertools.islice(find_all_subtitle_files(html_text), 10):
        pprint(sub_file, indent=2)


if __name__ == "__main__":
    main()



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
import asyncio
import time
import typing

from httpx import Headers


def parse_num(value: str) -> int | float:
    try:
        return int(value)
    except ValueError:
        return float(value)


def header_key_to_field_name(header_key: str) -> str:
    return header_key.replace("x-ratelimit-", "").replace("-", "_")


SLEEP_ENSURANCE_DELAY = 0.1


class RateLimit(typing.NamedTuple):
    # The number of requests that can be made.
    limit: int
    # How many requests are left before hitting a 429.
    remaining: int
    # The UNIX timestamp (seconds since midnight UTC on January 1st 1970) at which the rate limit resets.
    # This can have a fractional component for milliseconds.
    reset: int
    # The total time in seconds to wait for the rate limit to restart.
    # This can have a fractional component for milliseconds.
    reset_after: float | int = 0

    @classmethod
    def from_headers(cls, headers: Headers) -> typing.Self:
        return cls(
            **{
                header_key_to_field_name(key): parse_num(value)
                for key, value in headers.items()
                if key.startswith("x-ratelimit-")
            }
        )

    async def sleep(self):
        await asyncio.sleep(max(0, self.reset_after) + SLEEP_ENSURANCE_DELAY)


def main():
    headers = Headers(
        [
            ("content-type", "application/json"),
            ("access-control-allow-credentials", "true"),
            ("x-ratelimit-limit", "25"),
            ("x-ratelimit-remaining", "24"),
            ("x-ratelimit-reset", "1714518300"),
            ("vary", "origin, access-control-request-method, access-control-request-headers"),
            ("vary", "accept-encoding"),
            ("content-encoding", "gzip"),
            ("transfer-encoding", "chunked"),
            ("date", "Tue, 30 Apr 2024 23:04:52 GMT"),
        ]
    )
    print(RateLimit.from_headers(headers))


if __name__ == "__main__":
    main()



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import dataclasses
import datetime
import json
import pathlib
import typing
from collections.abc import Sequence
from pprint import pprint

from kitsunekko_tools.common import fs_name_strip


class ApiDirectoryFlagsDict(typing.TypedDict):
    adult: bool
    anime: bool
    external: bool
    low_quality: bool
    movie: bool


class ApiDirectoryDict(typing.TypedDict):
    id: int
    english_name: str
    flags: ApiDirectoryFlagsDict
    japanese_name: str
    last_modified: str
    name: str
    notes: typing.NotRequired[str]
    tmdb_id: typing.NotRequired[str]
    anilist_id: typing.NotRequired[int]
    creator_id: typing.NotRequired[int]


def describe_entry_type(flags: ApiDirectoryFlagsDict) -> str:
    return f'{"anime" if flags["anime"] else "drama"}_{"movie" if flags["movie"] else "tv"}'


def parse_api_time(time: str) -> datetime.datetime:
    return datetime.datetime.fromisoformat(time)


def format_api_time(time: datetime.datetime):
    return time.isoformat("T").replace("+00:00", "Z")


def nuke_key(d: dict, key: str) -> None:
    try:
        del d[key]
    except KeyError:
        pass


EntryId = typing.NewType("EntryId", int)


@dataclasses.dataclass(frozen=True)
class ApiDirectoryEntry:
    entry_id: EntryId  # used to query API for files in the directory
    name: str  # name of the anime and the directory on the disk.
    entry_type: str
    last_modified: datetime.datetime  # format RFC3339: '2024-04-27T17:54:01Z'
    english_name: str | None = None
    japanese_name: str | None = None
    anilist_id: int | None = None
    tmdb_id: str | None = None

    @classmethod
    def from_api_json(cls, json_dict: ApiDirectoryDict) -> typing.Self:
        """
        Construct self from the API json response.
        """
        return cls(
            entry_id=EntryId(json_dict["id"]),
            name=fs_name_strip(json_dict["name"]),
            entry_type=describe_entry_type(json_dict["flags"]),
            last_modified=parse_api_time(json_dict["last_modified"]),
            english_name=json_dict.get("english_name", "").strip(),
            japanese_name=json_dict.get("japanese_name", "").strip(),
            anilist_id=json_dict.get("anilist_id"),
            tmdb_id=json_dict.get("tmdb_id"),
        )

    def write_to_file(self, fp: typing.TextIO) -> None:
        """
        Format self and store on disk.
        The schema differs a bit from what the program receives from the remote server.
        """
        as_dict = dataclasses.asdict(self)
        as_dict["last_modified"] = format_api_time(self.last_modified)
        nuke_key(as_dict, "dir_path")  # no need to store this on disk
        json.dump(
            {k: v for k, v in as_dict.items() if v},
            fp,
            indent=2,
            ensure_ascii=False,
        )


@dataclasses.dataclass(frozen=True)
class KitsuDirectoryMeta(ApiDirectoryEntry):
    dir_path: pathlib.Path = pathlib.Path()

    @classmethod
    def from_local_file(cls, f: typing.TextIO, dir_path: pathlib.Path) -> typing.Self:
        return cls(**cls._load_json(f), dir_path=dir_path)

    @staticmethod
    def _load_json(f: typing.TextIO) -> dict:
        data = json.load(f)
        data["last_modified"] = parse_api_time(data["last_modified"])
        return data


def iter_catalog_directories(json_response: Sequence[ApiDirectoryDict]) -> typing.Iterable[ApiDirectoryEntry]:
    for item in json_response:
        yield ApiDirectoryEntry.from_api_json(item)


def main():
    example = {
        "id": 923,
        "name": "Yuru Yuri Nachuyachumi!",
        "flags": {"anime": True, "low_quality": False, "external": True, "movie": False, "adult": False},
        "last_modified": "2024-04-12T02:52:11Z",
        "anilist_id": 20625,
        "english_name": "YuruYuri Nachuyachumi!",
        "japanese_name": "ゆるゆり\u3000なちゅやちゅみ！",
    }
    pprint(ApiDirectoryEntry.from_api_json(example), indent=2)
    example = {
        "id": 3153,
        "name": "Wild Heroes",
        "flags": {"anime": False, "low_quality": False, "external": False, "movie": False, "adult": False},
        "last_modified": "2024-04-18T19:20:14Z",
        "creator_id": 1,
        "tmdb_id": "tv:86984",
        "english_name": "Wild Heroes",
        "japanese_name": "ワイルド・ヒーローズ",
    }
    pprint(ApiDirectoryEntry.from_api_json(example), indent=2)


if __name__ == "__main__":
    main()



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
import pathlib

from kitsunekko_tools.api_access.root_directory import EntryId, KitsuDirectoryMeta
from kitsunekko_tools.common import fs_name_strip
from kitsunekko_tools.config import KitsuConfig
from kitsunekko_tools.consts import IGNORE_FILENAME, INFO_FILENAME, TRASH_DIR_NAME

SKIP_FILES = (IGNORE_FILENAME, INFO_FILENAME, TRASH_DIR_NAME)


def move_files(old_dir: pathlib.Path, new_dir: pathlib.Path) -> None:
    for entry in old_dir.iterdir():
        if entry.is_dir():
            move_files(entry, new_dir / entry.name)
            continue
        if entry.name in SKIP_FILES:
            continue
        assert entry.is_file(), "entry must be a file."
        new_path = new_dir / entry.relative_to(old_dir)
        new_path.parent.mkdir(exist_ok=True)
        if new_path.exists():
            entry.unlink()
        else:
            entry.rename(new_path)
    nuke_dir(old_dir)


def nuke_dir(directory: pathlib.Path) -> None:
    (directory / INFO_FILENAME).unlink(missing_ok=True)
    directory.rmdir()


def rename_badly_named_directories(config: KitsuConfig) -> None:
    for directory in config.destination.iterdir():
        if directory.name in SKIP_FILES:
            continue
        sanitized_name = fs_name_strip(directory.name)
        if sanitized_name == directory.name:
            continue
        new_dir = directory.parent / sanitized_name
        print(f"moving '{directory}' to '{new_dir}'")
        if new_dir.exists():
            move_files(directory, new_dir)
        else:
            directory.rename(new_dir)


def merge_directories(config: KitsuConfig) -> None:
    id2master: dict[EntryId, KitsuDirectoryMeta] = {}
    name2id: dict[str, EntryId] = {}

    for directory in config.destination.iterdir():
        if directory.name in SKIP_FILES:
            continue
        try:
            with open(directory / INFO_FILENAME, encoding="utf-8") as f:
                meta = KitsuDirectoryMeta.from_local_file(f, dir_path=directory)
        except FileNotFoundError:
            continue

        if meta.entry_id not in id2master:
            id2master[meta.entry_id] = meta
        else:
            id2master[meta.entry_id] = meta = max(meta, id2master[meta.entry_id], key=lambda d: d.last_modified)

        name2id[directory.name.lower()] = meta.entry_id
        if meta.english_name:
            name2id[meta.english_name.lower()] = meta.entry_id
        if meta.japanese_name:
            name2id[meta.japanese_name.lower()] = meta.entry_id

    for directory in config.destination.iterdir():
        if directory.name in SKIP_FILES:
            continue
        try:
            master_entry = id2master[name2id[directory.name.lower()]]
        except KeyError:
            continue
        if master_entry.dir_path == directory:
            continue
        else:
            print(f"moving '{directory}' to '{master_entry.dir_path}'")
            move_files(directory, master_entry.dir_path)


def delete_empty_directories(config: KitsuConfig) -> None:
    for directory in config.destination.iterdir():
        if directory.name in SKIP_FILES:
            continue
        entries = [entry for entry in directory.iterdir() if entry.name not in SKIP_FILES]
        try:
            extra_files = [*directory.joinpath(TRASH_DIR_NAME).iterdir()]
        except FileNotFoundError:
            extra_files = []
        if not entries and not extra_files:
            print(f"deleting empty dir: {directory}")
            nuke_dir(directory)


def sanitize_directories(config: KitsuConfig) -> None:
    rename_badly_named_directories(config)
    merge_directories(config)
    delete_empty_directories(config)
